<?php

namespace Drupal\cambridge_migrations\Drush\Commands;

use Drush\Commands\DrushCommands;
use Drupal\Core\Entity\EntityTypeManagerInterface;
use Drupal\Core\Database\Connection;
use Drupal\Core\Config\ConfigFactoryInterface;
use Drupal\Core\File\FileSystemInterface;
use Drupal\cambridge_migrations\Utility\MediaEntityConverter;
use Symfony\Component\DependencyInjection\ContainerInterface;

/**
 * Drush command for post-migration tasks.
 */
class MigratePostTasksCommand extends DrushCommands {

  /**
   * The entity type manager.
   *
   * @var \Drupal\Core\Entity\EntityTypeManagerInterface
   */
  protected $entityTypeManager;

  /**
   * The file system service.
   *
   * @var \Drupal\Core\File\FileSystemInterface
   */
  protected $fileSystem;

  /**
   * Old (source) DB connection.
   *
   * @var \Drupal\Core\Database\Connection
   */
  protected $sourceDb;

  /**
   * New (target) DB connection (for migration map lookups, etc.).
   *
   * @var \Drupal\Core\Database\Connection
   */
  protected $targetDb;

  /**
   * The config factory.
   *
   * @var \Drupal\Core\Config\ConfigFactoryInterface
   */
  protected $configFactory;

  /**
   * The media entity converter service.
   *
   * @var \Drupal\cambridge_migrations\Utility\MediaEntityConverter
   */
  protected $mediaEntityConverter;

  /**
   * Constructs a new MigratePostTasksCommand object.
   */
  public function __construct(
    EntityTypeManagerInterface $entity_type_manager,
    FileSystemInterface $file_system,
    Connection $database,
    ConfigFactoryInterface $config_factory,
    MediaEntityConverter $media_entity_converter
  ) {
    parent::__construct();
    $this->entityTypeManager = $entity_type_manager;
    $this->fileSystem = $file_system;
    $this->configFactory = $config_factory;
    $this->mediaEntityConverter = $media_entity_converter;

    // "migrate" => old D7 DB; "default" => new D10 DB. Adjust as needed.
    $this->sourceDb = \Drupal\Core\Database\Database::getConnection('default', 'migrate');
    $this->targetDb = \Drupal\Core\Database\Database::getConnection('default', 'default');
  }

  /**
   * Performs post-migration tasks.
   *
   * This command performs several post-migration tasks:
   * 1. Updates all custom blocks to use the Basic HTML text format
   * 2. Fixes block regions to match their original D7 regions when possible
   * 3. Processes block content to convert D7 media entities to D10 media entities
   *
   * @command cambridge:post-migration-tasks
   * @aliases cm-post-tasks
   */
  public function postMigrationTasks() {
    $this->logger()->notice(dt('Starting post-migration tasks...'));
    
    // Process block content to convert D7 media entities to D10 media entities
    // This needs to be done BEFORE updating the text format
    $this->processBlockContent();
    
    // Update all custom blocks to use Basic HTML format
    $this->updateBlocksTextFormat();
    
    // Fix block regions
    $this->fixBlockRegions();
    
    // Clear all caches to ensure changes take effect
    $this->logger()->notice(dt('Clearing all caches...'));
    // Use drupal_flush_all_caches() to clear all caches
    \drupal_flush_all_caches();
    
    $this->logger()->notice(dt('Post-migration tasks completed.'));
  }

  /**
   * Updates all custom blocks to use the Basic HTML text format.
   */
  protected function updateBlocksTextFormat() {
    try {
      // Query the database directly to find all block content with body fields
      $query = $this->targetDb->select('block_content_field_data', 'b');
      $query->join('block_content__body', 'bb', 'b.id = bb.entity_id');
      $query->fields('b', ['id', 'type']);
      $query->fields('bb', ['body_format']);
      $query->condition('bb.body_format', 'basic_html', '!=');
      $results = $query->execute()->fetchAll();
      
      if (empty($results)) {
        $this->logger()->notice(dt('No blocks found that need format updating.'));
        return;
      }
      
      $updated_count = 0;
      $error_count = 0;
      
      foreach ($results as $result) {
        try {
          // Load the block
          $block = $this->entityTypeManager->getStorage('block_content')->load($result->id);
          if (!$block) {
            $this->logger()->warning(dt('Block @id not found.', [
              '@id' => $result->id,
            ]));
            continue;
          }
          
          // Update the format in the database directly
          $this->targetDb->update('block_content__body')
            ->fields(['body_format' => 'basic_html'])
            ->condition('entity_id', $result->id)
            ->execute();
            
          // Also update any revisions
          $this->targetDb->update('block_content_revision__body')
            ->fields(['body_format' => 'basic_html'])
            ->condition('entity_id', $result->id)
            ->execute();
          
          $updated_count++;
          $this->logger()->notice(dt('Updated block @id format from @old to basic_html.', [
            '@id' => $result->id,
            '@old' => $result->body_format,
          ]));
        }
        catch (\Exception $e) {
          $error_count++;
          $this->logger()->error(dt('Error updating block @id: @message', [
            '@id' => $result->id,
            '@message' => $e->getMessage(),
          ]));
        }
      }
      
      // Clear caches to ensure the changes take effect
      \Drupal::service('cache.render')->invalidateAll();
      
      $this->logger()->notice(dt('Block format update complete. Updated: @updated, Errors: @errors', [
        '@updated' => $updated_count,
        '@errors' => $error_count,
      ]));
    }
    catch (\Exception $e) {
      $this->logger()->error(dt('Block format update failed: @message', [
        '@message' => $e->getMessage(),
      ]));
    }
  }

  /**
   * Fixes block regions to match their original D7 regions when possible.
   */
  protected function fixBlockRegions() {
    try {
      $this->logger()->notice(dt('Starting block region fixes...'));
      
      // Get all blocks that were in the sidebar region in D7
      $d7_sidebar_blocks = $this->sourceDb->select('block', 'b')
        ->fields('b', ['module', 'delta', 'theme', 'region', 'visibility', 'pages', 'title'])
        ->condition('b.region', 'sidebar')
        ->condition('b.theme', 'cambridge_theme')
        ->condition('b.status', 1)
        ->execute()
        ->fetchAll();
      
      if (empty($d7_sidebar_blocks)) {
        $this->logger()->notice(dt('No sidebar blocks found in D7.'));
        return;
      }
      
      $updated_count = 0;
      $error_count = 0;
      
      foreach ($d7_sidebar_blocks as $d7_block) {
        try {
          // Find the corresponding D10 block
          $migration_map = $this->targetDb->select('migrate_map_upgrade_d7_block', 'm')
            ->fields('m', ['destid1'])
            ->condition('sourceid1', $d7_block->module)
            ->condition('sourceid2', $d7_block->delta)
            ->condition('sourceid3', $d7_block->theme)
            ->execute()
            ->fetchField();
          
          if (!$migration_map) {
            $this->logger()->info(dt('No migration mapping found for D7 block @module-@delta.', [
              '@module' => $d7_block->module,
              '@delta' => $d7_block->delta,
            ]));
            continue;
          }
          
          // Load the block configuration
          $block_config = $this->configFactory->getEditable('block.block.' . $migration_map);
          
          if (!$block_config || $block_config->isNew()) {
            $this->logger()->info(dt('Block configuration not found for @id.', [
              '@id' => $migration_map,
            ]));
            continue;
          }
          
          // Check if the block is in the content region
          $current_region = $block_config->get('region');
          if ($current_region !== 'content') {
            $this->logger()->info(dt('Block @id is already in region @region, not content.', [
              '@id' => $migration_map,
              '@region' => $current_region,
            ]));
            continue;
          }
          
          // Update the region to sidebar
          $block_config->set('region', 'sidebar');
          $block_config->save();
          
          $updated_count++;
          $this->logger()->notice(dt('Moved block @id from content to sidebar region.', [
            '@id' => $migration_map,
          ]));
        }
        catch (\Exception $e) {
          $error_count++;
          $this->logger()->error(dt('Error updating block region: @message', [
            '@message' => $e->getMessage(),
          ]));
        }
      }
      
      // Clear caches to ensure the changes take effect
      \Drupal::service('cache.render')->invalidateAll();
      
      $this->logger()->notice(dt('Block region fixes complete. Updated: @updated, Errors: @errors', [
        '@updated' => $updated_count,
        '@errors' => $error_count,
      ]));
    }
    catch (\Exception $e) {
      $this->logger()->error(dt('Block region fixes failed: @message', [
        '@message' => $e->getMessage(),
      ]));
    }
  }

  /**
   * Process block content to convert D7 media entities to D10 media entities.
   */
  protected function processBlockContent() {
    try {
      $this->logger()->notice(dt('Starting block content processing...'));
      
      // Query the database to find all block content with body fields
      $query = $this->targetDb->select('block_content_field_data', 'b');
      $query->join('block_content__body', 'bb', 'b.id = bb.entity_id');
      $query->fields('b', ['id', 'type']);
      $query->fields('bb', ['body_value', 'body_format', 'delta', 'langcode']);
      $results = $query->execute()->fetchAll();
      
      if (empty($results)) {
        $this->logger()->notice(dt('No blocks found that need content processing.'));
        return;
      }
      
      $updated_count = 0;
      $error_count = 0;
      
      foreach ($results as $result) {
        try {
          // Skip if no content
          if (empty($result->body_value)) {
            $this->logger()->notice(dt('Skipping block @id - empty content', [
              '@id' => $result->id,
            ]));
            continue;
          }
          
          // Log the original content for debugging
          $this->logger()->notice(dt('Original block @id content: @content', [
            '@id' => $result->id,
            '@content' => substr($result->body_value, 0, 60) . (strlen($result->body_value) > 60 ? '...' : ''),
          ]));
          
          // Check if the content contains any media tokens
          if (!preg_match('/\[\[\s*(\{.*?"fid":.*?\})\s*\]\]/s', $result->body_value)) {
            $this->logger()->notice(dt('Skipping block @id - no media tokens found', [
              '@id' => $result->id,
            ]));
            continue;
          }
          
          // Process the content to convert D7 media entities to D10 media entities
          $processed_content = $this->mediaEntityConverter->processContent(
            $result->body_value,
            $result->body_format,
            'basic_html'
          );
          
          // Skip if no changes were made
          if ($processed_content['value'] === $result->body_value) {
            $this->logger()->notice(dt('Skipping block @id - no changes made', [
              '@id' => $result->id,
            ]));
            continue;
          }
          
          // Log the processed content for debugging
          $this->logger()->notice(dt('Processed block @id content: @content', [
            '@id' => $result->id,
            '@content' => substr($processed_content['value'], 0, 60) . (strlen($processed_content['value']) > 60 ? '...' : ''),
          ]));
          
          // Update the block content in the database
          $update_count = $this->targetDb->update('block_content__body')
            ->fields([
              'body_value' => $processed_content['value'],
              'body_format' => $processed_content['format'],
            ])
            ->condition('entity_id', $result->id)
            ->condition('delta', $result->delta)
            ->condition('langcode', $result->langcode)
            ->execute();
            
          // Also update any revisions
          $revision_update_count = $this->targetDb->update('block_content_revision__body')
            ->fields([
              'body_value' => $processed_content['value'],
              'body_format' => $processed_content['format'],
            ])
            ->condition('entity_id', $result->id)
            ->condition('delta', $result->delta)
            ->condition('langcode', $result->langcode)
            ->execute();
          
          $updated_count++;
          $this->logger()->notice(dt('Updated block @id content with processed media entities. DB updates: @updates, Revision updates: @rev_updates', [
            '@id' => $result->id,
            '@updates' => $update_count,
            '@rev_updates' => $revision_update_count,
          ]));
          
          // Also update the entity directly to ensure it's properly saved
          try {
            $block = $this->entityTypeManager->getStorage('block_content')->load($result->id);
            if ($block && $block->hasField('body')) {
              $block->set('body', [
                'value' => $processed_content['value'],
                'format' => $processed_content['format'],
              ]);
              $block->save();
              $this->logger()->notice(dt('Successfully saved block entity @id with updated content', [
                '@id' => $result->id,
              ]));
            }
          }
          catch (\Exception $e) {
            $this->logger()->warning(dt('Could not update block entity @id directly: @message', [
              '@id' => $result->id,
              '@message' => $e->getMessage(),
            ]));
          }
        }
        catch (\Exception $e) {
          $error_count++;
          $this->logger()->error(dt('Error processing block @id content: @message', [
            '@id' => $result->id,
            '@message' => $e->getMessage(),
          ]));
        }
      }
      
      // Clear caches to ensure the changes take effect
      \Drupal::service('cache.render')->invalidateAll();
      
      $this->logger()->notice(dt('Block content processing complete. Updated: @updated, Errors: @errors', [
        '@updated' => $updated_count,
        '@errors' => $error_count,
      ]));
    }
    catch (\Exception $e) {
      $this->logger()->error(dt('Block content processing failed: @message', [
        '@message' => $e->getMessage(),
      ]));
    }
  }

  /**
   * {@inheritdoc}
   */
  public static function create(ContainerInterface $container) {
    return new static(
      $container->get('entity_type.manager'),
      $container->get('file_system'),
      $container->get('database'),
      $container->get('config.factory'),
      $container->get('cambridge_migrations.media_entity_converter')
    );
  }
}
