<?php

namespace Drupal\cambridge_migrations\Drush\Commands;

use Drush\Commands\DrushCommands;
use Drupal\Core\Entity\EntityTypeManagerInterface;
use Drupal\Core\File\FileSystemInterface;
use Drupal\Core\Database\Connection;
use Drupal\paragraphs\Entity\Paragraph;
use Drupal\cambridge_migrations\Utility\MediaEntityConverter;
use Symfony\Component\DependencyInjection\ContainerInterface;

/**
 * Drush command for migrating text blocks to text paragraphs.
 */
class MigrateTextBlocksCommand extends DrushCommands {

  /**
   * The entity type manager.
   *
   * @var \Drupal\Core\Entity\EntityTypeManagerInterface
   */
  protected $entityTypeManager;

  /**
   * The file system service.
   *
   * @var \Drupal\Core\File\FileSystemInterface
   */
  protected $fileSystem;

  /**
   * Old (source) DB connection.
   *
   * @var \Drupal\Core\Database\Connection
   */
  protected $sourceDb;

  /**
   * New (target) DB connection (for migration map lookups, etc.).
   *
   * @var \Drupal\Core\Database\Connection
   */
  protected $targetDb;

  /**
   * The media entity converter service.
   *
   * @var \Drupal\cambridge_migrations\Utility\MediaEntityConverter
   */
  protected $mediaEntityConverter;

  /**
   * Constructs a new MigrateTextBlocksCommand object.
   */
  public function __construct(
    EntityTypeManagerInterface $entity_type_manager,
    FileSystemInterface $file_system,
    Connection $database,
    MediaEntityConverter $media_entity_converter
  ) {
    parent::__construct();
    $this->entityTypeManager = $entity_type_manager;
    $this->fileSystem = $file_system;
    $this->mediaEntityConverter = $media_entity_converter;

    // "migrate" => old D7 DB; "default" => new D10 DB. Adjust as needed.
    $this->sourceDb = \Drupal\Core\Database\Database::getConnection('default', 'migrate');
    $this->targetDb = \Drupal\Core\Database\Database::getConnection('default', 'default');
  }


  /**
   * Remove content teaser paragraphs from a node.
   */
  protected function removeContentTeasers($node) {
    if (!$node->hasField('field_paragraph')) {
      return;
    }

    $paragraphs = $node->get('field_paragraph')->referencedEntities();
    $updated_paragraphs = [];
    $removed_count = 0;

    foreach ($paragraphs as $paragraph) {
      if ($paragraph->getType() !== 'content_teaser') {
        $updated_paragraphs[] = [
          'target_id' => $paragraph->id(),
          'target_revision_id' => $paragraph->getRevisionId(),
        ];
      }
      else {
        $paragraph->delete();
        $removed_count++;
      }
    }

    if ($removed_count > 0) {
      $node->set('field_paragraph', $updated_paragraphs);
      $node->save();
      $this->logger()->notice(dt('Removed @count content teaser paragraph(s) from node @nid', [
        '@count' => $removed_count,
        '@nid' => $node->id(),
      ]));
    }
  }

  /**
   * Remove all paragraphs from a node's sidebar items field.
   */
  protected function removeSidebarItems($node) {
    if (!$node->hasField('field_sidebar_items')) {
      return;
    }

    $paragraphs = $node->get('field_sidebar_items')->referencedEntities();
    $removed_count = 0;

    foreach ($paragraphs as $paragraph) {
      $paragraph->delete();
      $removed_count++;
    }

    if ($removed_count > 0) {
      $node->set('field_sidebar_items', []);
      $node->save();
      $this->logger()->notice(dt('Removed @count sidebar paragraph(s) from node @nid', [
        '@count' => $removed_count,
        '@nid' => $node->id(),
      ]));
    }
  }


  /**
   * Process text block content to convert media tokens and clean up.
   *
   * @param string $content
   *   The raw content from D7.
   * @param string $format
   *   The text format to use.
   *
   * @return array
   *   An array with 'value' and 'format' keys for the processed content.
   */
  protected function processTextBlockContent($content, $format) {
    return $this->mediaEntityConverter->processContent($content, $format, 'filtered_html');
  }

  /**
   * Migrate text blocks to text paragraphs.
   *
   * @command cambridge:migrate-text-blocks
   * @aliases cm-text-blocks
   */
  public function migrateTextBlocks() {
    try {
      // First, migrate main content text blocks
      $this->migrateMainContentTextBlocks();
      
      // Rebuild caches between migrations
      drupal_flush_all_caches();

      // Then, create sidebar paragraph items
      $sidebar_items = $this->createSidebarParagraphItems();
      
      // Finally, insert sidebar items into the database
      $this->insertSidebarItemsIntoDatabase($sidebar_items);
      
      // Rebuild caches again
      drupal_flush_all_caches();

      $this->logger()->notice(dt('Text block migration complete for both main content and sidebar items.'));
    }
    catch (\Exception $e) {
      $this->logger()->error(dt('Migration failed: @message', [
        '@message' => $e->getMessage(),
      ]));
    }
  }

  /**
   * Migrate text blocks to text paragraphs in the main content field.
   */
  protected function migrateMainContentTextBlocks() {
    try {
      // Query the D7 database for text blocks in the main content field.
      $query = $this->sourceDb->select('field_data_field_content_items', 'n')
        ->fields('n', ['entity_id', 'field_content_items_value']);

      // Join with paragraphs_item to get only text blocks
      $query->join(
        'paragraphs_item',
        'p',
        'p.item_id = n.field_content_items_value AND p.bundle = :bundle',
        [':bundle' => 'text_block']
      );
      $query->fields('p', ['item_id', 'bundle', 'field_name']);

      // Join with field_data_field_paragraph_heading for headings.
      $query->leftJoin(
        'field_data_field_paragraph_heading',
        'h',
        'p.item_id = h.entity_id AND h.entity_type = :entity_type',
        [':entity_type' => 'paragraphs_item']
      );
      $query->fields('h', ['field_paragraph_heading_value']);

      // Join with field_data_field_paragraph_text_content for body + format.
      $query->leftJoin(
        'field_data_field_paragraph_text_content',
        't',
        'p.item_id = t.entity_id AND t.entity_type = :entity_type',
        [':entity_type' => 'paragraphs_item']
      );
      $query->fields('t', ['field_paragraph_text_content_value', 'field_paragraph_text_content_format']);

      $text_blocks = $query->execute()->fetchAll();
      if (empty($text_blocks)) {
        $this->logger()->warning(dt('No main content text blocks found to migrate.'));
        return;
      }

      $success_count = 0;
      $error_count = 0;

      foreach ($text_blocks as $block) {
        try {
          // Skip if no node reference.
          if (empty($block->entity_id)) {
            $this->logger()->warning(dt('Skipping text block @id - no node reference found', [
              '@id' => $block->item_id,
            ]));
            continue;
          }

          // Combine heading + body text.
          $content = '';
          if (!empty($block->field_paragraph_heading_value)) {
            $content .= '<h2>' . $block->field_paragraph_heading_value . '</h2>';
          }
          if (!empty($block->field_paragraph_text_content_value)) {
            $content .= $block->field_paragraph_text_content_value;
          }

          // Process the content
          $processed_content = $this->processTextBlockContent(
            $content,
            $block->field_paragraph_text_content_format
          );

          // Load the corresponding D10 node.
          $node = $this->entityTypeManager->getStorage('node')->load($block->entity_id);
          if (!$node) {
            throw new \Exception("Node {$block->entity_id} not found in D10.");
          }

          // Remove content teasers before adding this new paragraph.
          $this->removeContentTeasers($node);

          // Create a new paragraph of type 'text'.
          $paragraph = Paragraph::create([
            'type' => 'text',
            'langcode' => 'und', // Set paragraph language to 'und' (undefined) to match existing data
            'field_text' => [
              'value' => $processed_content['value'],
              'format' => $processed_content['format'], // Use the imported format for main content
            ],
          ]);
          $paragraph->save();

          // Attach the new paragraph to the node's field_paragraph.
          $node->field_paragraph[] = [
            'target_id' => $paragraph->id(),
            'target_revision_id' => $paragraph->getRevisionId(),
          ];
          $node->save();
          
          // Debug: Log the structure of the main content field after saving
          $this->logger()->debug(dt('Main content field structure after save for node @nid: @count paragraphs', [
            '@nid' => $node->id(),
            '@count' => count($node->field_paragraph),
          ]));

          $success_count++;
          $this->logger()->notice(dt('Successfully migrated main content text block @id to node @nid', [
            '@id' => $block->item_id,
            '@nid' => $block->entity_id,
          ]));
        }
        catch (\Exception $e) {
          $error_count++;
          $this->logger()->error(dt('Error migrating main content text block @id: @message', [
            '@id' => $block->item_id,
            '@message' => $e->getMessage(),
          ]));
        }
      }

      $this->logger()->notice(dt('Main content text block migration complete. Successes: @success, Errors: @errors', [
        '@success' => $success_count,
        '@errors' => $error_count,
      ]));
    }
    catch (\Exception $e) {
      $this->logger()->error(dt('Main content text block migration failed: @message', [
        '@message' => $e->getMessage(),
      ]));
    }
  }

  /**
   * Create sidebar paragraph items and return an array of data for database insertion.
   * 
   * @return array
   *   An array of sidebar item data for database insertion.
   */
  protected function createSidebarParagraphItems() {
    try {
      $sidebar_items = [];
      
      // Query the D7 database for text blocks in the sidebar field.
      $query = $this->sourceDb->select('field_data_field_sidebar_items', 'n')
        ->fields('n', ['entity_id', 'field_sidebar_items_value']);

      // Join with paragraphs_item to get only text blocks
      $query->join(
        'paragraphs_item',
        'p',
        'p.item_id = n.field_sidebar_items_value AND p.bundle = :bundle',
        [':bundle' => 'text_block']
      );
      $query->fields('p', ['item_id', 'bundle', 'field_name']);

      // Join with field_data_field_paragraph_heading for headings.
      $query->leftJoin(
        'field_data_field_paragraph_heading',
        'h',
        'p.item_id = h.entity_id AND h.entity_type = :entity_type',
        [':entity_type' => 'paragraphs_item']
      );
      $query->fields('h', ['field_paragraph_heading_value']);

      // Join with field_data_field_paragraph_text_content for body + format.
      $query->leftJoin(
        'field_data_field_paragraph_text_content',
        't',
        'p.item_id = t.entity_id AND t.entity_type = :entity_type',
        [':entity_type' => 'paragraphs_item']
      );
      $query->fields('t', ['field_paragraph_text_content_value', 'field_paragraph_text_content_format']);

      $text_blocks = $query->execute()->fetchAll();
      if (empty($text_blocks)) {
        $this->logger()->warning(dt('No sidebar text blocks found to migrate.'));
        return $sidebar_items;
      }

      $success_count = 0;
      $error_count = 0;
      $processed_nodes = [];

      foreach ($text_blocks as $block) {
        try {
          // Skip if no node reference.
          if (empty($block->entity_id)) {
            $this->logger()->warning(dt('Skipping sidebar text block @id - no node reference found', [
              '@id' => $block->item_id,
            ]));
            continue;
          }

          // Load the corresponding D10 node.
          $node = $this->entityTypeManager->getStorage('node')->load($block->entity_id);
          if (!$node) {
            throw new \Exception("Node {$block->entity_id} not found in D10.");
          }

          // Remove existing sidebar items before adding new ones, but only once per node
          if (!in_array($block->entity_id, $processed_nodes)) {
            $this->removeSidebarItems($node);
            $processed_nodes[] = $block->entity_id;
          }

          // Combine heading + body text.
          $content = '';
          if (!empty($block->field_paragraph_heading_value)) {
            $content .= '<h2>' . $block->field_paragraph_heading_value . '</h2>';
          }
          if (!empty($block->field_paragraph_text_content_value)) {
            $content .= $block->field_paragraph_text_content_value;
          }

          // Process the content
          $processed_content = $this->processTextBlockContent(
            $content,
            $block->field_paragraph_text_content_format
          );

          // Create a new paragraph of type 'text'.
          $paragraph = Paragraph::create([
            'type' => 'text',
            'langcode' => 'und', // Set paragraph language to 'und' (undefined) to match existing data
            'field_text' => [
              'value' => $processed_content['value'],
              'format' => 'basic_html', // Explicitly set to basic_html as required
            ],
          ]);
          $paragraph->save();

          // Store the paragraph and node information for later database insertion
          // Get the node revision ID safely - in Drupal 10, we can use the vid field
          $node_revision_id = $node->id(); // Default to node ID if revision ID not available
          if ($node->hasField('vid') && !$node->get('vid')->isEmpty()) {
            $node_revision_id = $node->get('vid')->value;
          }
          
          $sidebar_items[] = [
            'node_id' => $node->id(),
            'node_revision_id' => $node_revision_id,
            'node_bundle' => $node->bundle(),
            'paragraph_id' => $paragraph->id(),
            'paragraph_revision_id' => $paragraph->getRevisionId(),
          ];

          // Debug: Log detailed information about the paragraph and the node reference
          $this->logger()->notice(dt('IMPORTANT: Sidebar paragraph created for node @nid: ID=@pid, RevID=@revid, Type=@type, Lang=@lang, Format=@format', [
            '@nid' => $node->id(),
            '@pid' => $paragraph->id(),
            '@revid' => $paragraph->getRevisionId(),
            '@type' => $paragraph->getType(),
            '@lang' => $paragraph->language()->getId(),
            '@format' => $paragraph->get('field_text')->format,
          ]));

          $success_count++;
          $this->logger()->notice(dt('Successfully created sidebar paragraph for text block @id to node @nid', [
            '@id' => $block->item_id,
            '@nid' => $block->entity_id,
          ]));
        }
        catch (\Exception $e) {
          $error_count++;
          $this->logger()->error(dt('Error creating sidebar paragraph for text block @id: @message', [
            '@id' => $block->item_id,
            '@message' => $e->getMessage(),
          ]));
        }
      }

      $this->logger()->notice(dt('Sidebar paragraph creation complete. Successes: @success, Errors: @errors', [
        '@success' => $success_count,
        '@errors' => $error_count,
      ]));
      
      return $sidebar_items;
    }
    catch (\Exception $e) {
      $this->logger()->error(dt('Sidebar paragraph creation failed: @message', [
        '@message' => $e->getMessage(),
      ]));
      return [];
    }
  }

  /**
   * Insert sidebar items into the database.
   * 
   * @param array $sidebar_items
   *   An array of sidebar item data for database insertion.
   */
  protected function insertSidebarItemsIntoDatabase(array $sidebar_items) {
    try {
      // Truncate the sidebar items tables to start afresh
      $this->targetDb->truncate('node__field_sidebar_items')->execute();
      $this->targetDb->truncate('node_revision__field_sidebar_items')->execute();
      $this->logger()->notice(dt('Truncated sidebar items tables to start afresh.'));
      
      if (empty($sidebar_items)) {
        $this->logger()->warning(dt('No sidebar items to insert into the database.'));
        return;
      }
      
      $success_count = 0;
      $error_count = 0;
      
      foreach ($sidebar_items as $item) {
        try {
          $node = $this->entityTypeManager->getStorage('node')->load($item['node_id']);
          $node->field_sidebar_items[] = [
            'target_id' => $item['paragraph_id'],
            'target_revision_id' => $item['paragraph_revision_id'],
          ];
          $node->save();

          $success_count++;
          $this->logger()->notice(dt('Successfully inserted sidebar item for node @nid', [
            '@nid' => $item['node_id'],
          ]));
        }
        catch (\Exception $e) {
          $error_count++;
          $this->logger()->error(dt('Error inserting sidebar item for node @nid: @message', [
            '@nid' => $item['node_id'],
            '@message' => $e->getMessage(),
          ]));
        }
      }
      
      $this->logger()->notice(dt('Sidebar item database insertion complete. Successes: @success, Errors: @errors', [
        '@success' => $success_count,
        '@errors' => $error_count,
      ]));
    }
    catch (\Exception $e) {
      $this->logger()->error(dt('Sidebar item database insertion failed: @message', [
        '@message' => $e->getMessage(),
      ]));
    }
  }

  /**
   * {@inheritdoc}
   */
  public static function create(ContainerInterface $container) {
    return new static(
      $container->get('entity_type_manager'),
      $container->get('file_system'),
      $container->get('database'),
      $container->get('cambridge_migrations.media_entity_converter')
    );
  }

}
