<?php

namespace Drupal\cambridge_migrations\Utility;

use Drupal\Core\Database\Connection;
use Drupal\Core\Entity\EntityTypeManagerInterface;
use Drupal\Core\Logger\LoggerChannelFactoryInterface;

/**
 * Utility class for converting D7 media entities to D10 media entities.
 */
class MediaEntityConverter {

  /**
   * The entity type manager.
   *
   * @var \Drupal\Core\Entity\EntityTypeManagerInterface
   */
  protected $entityTypeManager;

  /**
   * New (target) DB connection (for migration map lookups, etc.).
   *
   * @var \Drupal\Core\Database\Connection
   */
  protected $targetDb;

  /**
   * The logger channel.
   *
   * @var \Drupal\Core\Logger\LoggerChannelInterface
   */
  protected $logger;

  /**
   * Constructs a new MediaEntityConverter object.
   *
   * @param \Drupal\Core\Entity\EntityTypeManagerInterface $entity_type_manager
   *   The entity type manager.
   * @param \Drupal\Core\Database\Connection $database
   *   The database connection.
   * @param \Drupal\Core\Logger\LoggerChannelFactoryInterface $logger_factory
   *   The logger factory.
   */
  public function __construct(
    EntityTypeManagerInterface $entity_type_manager,
    Connection $database,
    LoggerChannelFactoryInterface $logger_factory
  ) {
    $this->entityTypeManager = $entity_type_manager;
    $this->targetDb = $database;
    $this->logger = $logger_factory->get('cambridge_migrations');
  }

  /**
   * Clean up HTML content by removing unwanted elements.
   *
   * @param string $content
   *   The content to clean.
   *
   * @return string
   *   The cleaned content.
   */
  public function cleanContent($content) {
    if (empty($content)) {
      return $content;
    }
    // Remove empty <p> tags.
    $content = preg_replace('/<p>\s*<\/p>/', '', $content);
    // Collapse multiple blank lines.
    $content = preg_replace('/(\r?\n){2,}/', "\n\n", $content);
    return trim($content);
  }

  /**
   * Helper to find the new D10 Media entity that corresponds to a given old D7 file fid.
   *
   * @param int $old_fid
   *   The old D7 file fid as stored in the token JSON.
   *
   * @return \Drupal\media\MediaInterface|null
   *   The mapped Media entity in the new site, or NULL if not found.
   */
  public function getMediaEntityFromOldFid($old_fid) {
    // Log the FID we're trying to map
    $this->logger->notice(sprintf('Attempting to map D7 file ID: %s', $old_fid));
    
    // Try multiple migration map tables
    $migration_map_tables = [
      'm_map_upgrade_d7_file_entity_image_public',
      'migrate_map_upgrade_d7_file_entity_image_public',
      'migrate_map_upgrade_d7_file',
      'migrate_map_upgrade_d7_file_entity_image',
    ];
    
    $mapped_revision_id = NULL;
    $used_table = NULL;
    
    foreach ($migration_map_tables as $table) {
      // Check if the table exists
      try {
        if (!$this->targetDb->schema()->tableExists($table)) {
          $this->logger->notice(sprintf('Migration map table %s does not exist', $table));
          continue;
        }
        
        // Look up "destid1" in the migration map table for this old_fid
        $query = $this->targetDb->select($table, 'm')
          ->fields('m', ['destid1'])
          ->condition('sourceid1', $old_fid);
        
        $this->logger->notice(sprintf('Querying table %s for sourceid1=%s: %s', 
          $table, $old_fid, $query->__toString()));
        
        $mapped_id = $query->execute()->fetchField();
        
        if ($mapped_id) {
          $mapped_revision_id = $mapped_id;
          $used_table = $table;
          $this->logger->notice(sprintf('Found mapping in table %s: %s -> %s', 
            $table, $old_fid, $mapped_revision_id));
          break;
        }
      }
      catch (\Exception $e) {
        $this->logger->error(sprintf('Error querying migration map table %s: %s', 
          $table, $e->getMessage()));
      }
    }

    if (!$mapped_revision_id) {
      // No mapping found in any table
      $this->logger->warning(sprintf('No mapping found for D7 file ID %s in any migration map table', $old_fid));
      return NULL;
    }

    // Try to load the media entity
    try {
      // First try loading as a revision
      $media = $this->entityTypeManager
        ->getStorage('media')
        ->loadRevision($mapped_revision_id);
      
      if ($media) {
        $this->logger->notice(sprintf('Successfully loaded media revision %s for D7 file ID %s', 
          $mapped_revision_id, $old_fid));
        return $media;
      }
      
      // If that fails, try loading as a regular entity
      $media = $this->entityTypeManager
        ->getStorage('media')
        ->load($mapped_revision_id);
      
      if ($media) {
        $this->logger->notice(sprintf('Successfully loaded media entity %s for D7 file ID %s', 
          $mapped_revision_id, $old_fid));
        return $media;
      }
      
      $this->logger->warning(sprintf('Could not load media entity or revision %s for D7 file ID %s', 
        $mapped_revision_id, $old_fid));
      return NULL;
    }
    catch (\Exception $e) {
      $this->logger->error(sprintf('Error loading media entity %s: %s', 
        $mapped_revision_id, $e->getMessage()));
      return NULL;
    }
  }

  /**
   * Checks if a requested media view mode actually exists on Drupal 10.
   * If not, returns "media_library" as a safe default.
   *
   * @param string $requested_view_mode
   *   The name of the requested view mode, e.g. "wysiwyg" or "media_library".
   *
   * @return string
   *   The validated view mode name, or "media_library" if not found.
   */
  public function getValidMediaViewMode($requested_view_mode) {
    // If none was requested, default to "media_library".
    if (empty($requested_view_mode)) {
      return 'media_library';
    }

    // Check whether the requested view mode is defined for "media" entities.
    $display_repository = \Drupal::service('entity_display.repository');
    $all_view_modes = $display_repository->getViewModes('media');
    // $all_view_modes is typically an array of [machine_name => Label].
    if (!array_key_exists($requested_view_mode, $all_view_modes)) {
      // Requested mode not found, fall back to "media_library".
      return 'media_library';
    }
    return $requested_view_mode;
  }

  /**
   * Process content to convert media tokens and clean up.
   *
   * @param string $content
   *   The raw content from D7.
   * @param string $format
   *   The text format to use.
   * @param string $default_format
   *   The default text format to use if none is provided.
   *
   * @return array
   *   An array with 'value' and 'format' keys for the processed content.
   */
  public function processContent($content, $format, $default_format = 'filtered_html') {
    // Log the content we're processing
    $this->logger->notice(sprintf('Processing content with format %s: %s', 
      $format, substr($content, 0, 100) . (strlen($content) > 100 ? '...' : '')));
    
    // Clean up content slightly.
    $content = $this->cleanContent($content);

    // Remove literal occurrences of 'text_block'.
    $content = str_replace('text_block', '', $content);

    // Determine the text format to use in D10: if not found, fall back to default_format
    $content_format = !empty($format) ? $format : $default_format;

    // Check if the content contains any media tokens
    if (preg_match('/\[\[\s*(\{.*?"fid":.*?\})\s*\]\]/s', $content)) {
      $this->logger->notice('Content contains media tokens, attempting to convert...');
    } else {
      $this->logger->notice('Content does not contain any media tokens.');
    }

    // Convert D7 media tokens to <drupal-media>.
    $content = preg_replace_callback(
      '/\[\[\s*(\{.*?"fid":.*?\})\s*\]\]/s',
      function ($matches) {
        $json_string = $matches[1];
        $truncated_json = substr($json_string, 0, 60) . (strlen($json_string) > 60 ? '...' : '');
        $this->logger->notice(sprintf('Found media token: %s', $truncated_json));
        
        $embed_data = json_decode($json_string, TRUE);
        if (is_array($embed_data) && isset($embed_data['fid'])) {
          // Get the correct new media entity.
          $old_fid = $embed_data['fid'];
          $this->logger->notice(sprintf('Extracted fid: %s', $old_fid));
          
          $media = $this->getMediaEntityFromOldFid($old_fid);
          if ($media) {
            // If "fields.format" is set in the old JSON, treat that as the requested view mode.
            // Otherwise default to "media_library".
            $requested_mode = !empty($embed_data['fields']['format'])
              ? $embed_data['fields']['format']
              : 'media_library';

            // Validate if this view mode actually exists in D10; fallback if not.
            $view_mode = $this->getValidMediaViewMode($requested_mode);

            $uuid = $media->uuid();
            $drupal_media_tag = '<drupal-media data-entity-type="media" data-entity-uuid="' . $uuid . '" data-view-mode="' . $view_mode . '"></drupal-media>';
            $this->logger->notice(sprintf('Converted to drupal-media tag: %s', $drupal_media_tag));
            return $drupal_media_tag;
          }
          else {
            // No mapping => log warning & keep original token.
            $this->logger->warning(sprintf('No media mapping found for old fid "%s".', $old_fid));
            return $matches[0];
          }
        }
        else {
          $this->logger->warning(sprintf('Failed to parse JSON or no fid found: %s', $json_string));
        }
        // If JSON parse fails or no "fid", leave token as-is.
        return $matches[0];
      },
      $content
    );

    // Log the processed content
    $this->logger->notice(sprintf('Processed content: %s', 
      substr($content, 0, 60) . (strlen($content) > 60 ? '...' : '')));

    return [
      'value' => $content,
      'format' => $content_format,
    ];
  }

}
